//
// Title:   FrequencyTransformationInterface.h 
// Created: November 28, 2002
// Authors: Daniel Cotting
//
// Copyright (c) 2001, 2002, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.
//
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.

#ifndef _FREQUENCYTRANSFORMATIONINTERFACE_H_EA6C0EE0_BC30_11d5_B625_A58C4DF45B22_INCLUDED
#define _FREQUENCYTRANSFORMATIONINTERFACE_H_EA6C0EE0_BC30_11d5_B625_A58C4DF45B22_INCLUDED

#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_eigen.h>

#include "../../../Core/Interfaces/src/SurfelInterface.h"
#include "WatermarkInterface.h"
#include "ModuleInterface.h"
#include "GenericVector.h"
#include "MySurfelList.h"

/**
 * This template computes a frequency transformation on a surfel list.
 *
 * Template parameters: <tt>nofAttributes</tt>: The number of attributes to embed watermark into.
 *
 * Usually the number of attributes is set to 3. If it is set to a value larger than 5
 * both positions and colors are considered.  To embed information only into the color 
 * attributes set the number to 3 and pass a color watermark to the constructor.
 */
template <unsigned nofAttributes>
class FrequencyTransformationInterface: public ModuleInterface
{

public:

	/**
	 * List containing spectral components.
	 */
	class SpectralList: public QList<GenericVector<nofAttributes> > {
	public:
		~SpectralList() {
			for (GenericVector<nofAttributes>* v = this->first(); v!=0; v = this->current()) {
				this->remove(v);
				assert(v!=NULL);
				delete v;
				v=NULL;
			}
			this->clear();
		}
	};

	/** 
	 * Encapsulates temporary data which can be passed from the forward transformation 
	 * to the backward transformation.
	 */
	class DataPack {};

	/**
	 * Performs the forward transformation.
	 *
	 * @param surfelList The surfels to transform into frequency domain.
	 * @param spectralListOut The <tt>SpectralList</tt> object which will store the result.
	 * @param dataPackOut The <tt>DataPack</tt> object which will store temporary data required for backward transformation.
	 */
	virtual void forwardTransformation(MySurfelList* surfelList,
								SpectralList** spectralListOut, 
								DataPack** dataPackOut) = 0;

	/**
	 * Performs a fast version of the forward transformation using the laplace matrix (does skip some
	 * steps not necessary if the topology of the points is the same as in a previous calculation).
	 *
	 * @param surfelList		The surfels to transform into frequency domain.
	 * @param dataPackIn		The <tt>DataPack</tt> object which already stores some information 
	 *							about a calculation performed on a point set with the same topology.
	 * @param spectralListOut	The <tt>SpectralList</tt> object which will store the result.
	 */
	virtual void forwardTransformation(
								MySurfelList* surfelList, 
								DataPack* dataPackIn,
								SpectralList** spectralListOut) = 0;

	/**
	 * Performs the backward transformation.
	 *
	 * @param surfelList The list of the surfels whose positions will be changed to the values resulting from backward transformation.
	 * @param spectralList The <tt>SpectralList</tt> object containing the spectral respresentation.
	 * @param dataPack The <tt>DataPack</tt> object storing temporary data required for backward transformation.
	 */
	virtual void backwardTransformation(MySurfelList* surfelList, 
								SpectralList* spectralList,
								DataPack* dataPack) = 0;

	/**
	 * Returns the name of this module.
	 *
	 * @return The name of this module.
	 */
	virtual const char* getModuleName() const = 0;

	/**
	 * Returns the description of this module.
	 *
	 * @return The description of this module.
	 */
	virtual const char* getModuleDescription() const = 0;
};

#endif // _FREQUENCYTRANSFORMATIONINTERFACE_H_EA6C0EE0_BC30_11d5_B625_A58C4DF45B22_INCLUDED

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
