//
// Title:   MajorityWatermark.cpp
// Created: November 28, 2002
// Authors: Daniel Cotting
//
// Copyright (c) 2001, 2002, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.
//
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

/**
 * @file MajorityWatermark.cpp
 * 
 * Watermark with error correction capabilities by using
 * majority voting.
 *
 * @author Daniel Cotting
 */

#include <assert.h>
#include "MajorityWatermark.h"

/**
 * Constructor for the majority watermark. 
 * The capacity can be varied, also the chip-rate (number of repetitions of 
 * every encoded bit, for error correction) and usage-divider (portion
 * of the frequencies used for embedding watermark).
 * Furthermore a strength for the watermark can be specified.
 * Note that although usage-divider has been introduced for frequency 
 * based embedding, this watermarking class is not restricted to these
 * frequency based approaches. Usage-divider just tells the watermarking
 * algorithm to reserve more frequencies than required: In the end only
 * lower frequencies are used to embed the watermark.
 * Chip-rate is internally three times higher because we embed the same bit into 
 * the x, y and z coordinate.
 *
 * @param capacity The length of the watermark to be used (possible values range from 1 to 160).
 * @param watermarkStrength The strength of the watermark.
 * @param watermarkChipRate The chip-rate of the watermark.
 * @param watermarkUsageDivider The usage divider of the watermark.
 * @param colorWatermarkIn Indicator whether the watermark has to be applied on color or position.
 * @param useFastVersionIn Indicator whether the watermark has to be applied using the fast embedding algorithm.
 */
MajorityWatermark::MajorityWatermark(unsigned int capacity, float watermarkStrength,
	unsigned int watermarkChipRate, unsigned int watermarkUsageDivider, 
	bool colorWatermarkIn, bool useFastVersionIn): 
	chipRate(watermarkChipRate),usageDivider(watermarkUsageDivider),
	strength(watermarkStrength),capacityInBits(capacity), 
	colorWatermark(colorWatermarkIn), useFastVersion(useFastVersionIn) {

	// our watermark consists of an array of char
	// we require the char data type to be at least 1 byte long
	int sizeOfUnsignedChar = sizeof(unsigned char);
	assert(sizeOfUnsignedChar>=1);

	// create the watermark array for the specified capacity...
	// size is sometimes too large, but this is not critical, so let's keep it that way.
	watermarkBits = new unsigned char[(getSizeInBits()+7)/8];

	// clear the whole watermark before using it
	for (unsigned int i=0; i<(getSizeInBits()+7)/8; i++) {
		watermarkBits[i] = 0;
	}
}

/** 
 * Destructor for the majority watermark.
 *
 * Cleans up dynamic data structures.
 */
MajorityWatermark::~MajorityWatermark() {
	delete[] watermarkBits;
}

/** 
 * Returns the total size in bits of this watermark.
 *
 * @return The total size in bits of this watermark.
 */
const unsigned int MajorityWatermark::getSizeInBits() const {
	return chipRate*usageDivider*capacityInBits;
} 

/**
 * Returns the size in bits of the payload of this watermark.
 * This is the number of bits that has to be encoded eventually.
 * 
 * @return The size in bits of the payload of this watermark.
 */
const unsigned int MajorityWatermark::getUsedSizeInBits() const {
	return chipRate*capacityInBits;
} 

/** 
 * Returns the number of user-defined bits which this watermark can hold.
 *
 * @return The number of user-defined bits which this watermark can hold.
 */
const unsigned int MajorityWatermark::getCapacityInBits() const {
	return capacityInBits;
} 

/** 
 * Returns the number of repetitions of every encoded bit. Repetition
 * is used for error correction.
 *
 * @return The number of repetitions of every encoded bit.
 */
const unsigned int MajorityWatermark::getChipRate() const {
	return chipRate;
} 

/**
 * Returns the fraction of the total size that is used to really encode
 * any data, e.g.\ a value of 2 means that the watermark reserves a 
 * total size 2 times as large as the payload size (this makes sense
 * to get more frequencies and only use the lower ones for encoding. 
 *
 * @return	The fraction of the total size that is used to really encode
 *			any data.
 */
const unsigned int MajorityWatermark::getUsageDivider() const {
	return usageDivider;
} 

/** 
 * Returns the strength of this watermark.
 *
 * @return The strength of this watermark.
 */
const float MajorityWatermark::getStrength() const {
	return strength;
} 

/** 
 * Returns whether the watermark is a color watermark or a position watermark.
 *
 * @return Whether the watermark is a color watermark or a position watermark.
 */
const bool MajorityWatermark::isColorWatermark() const {
	return colorWatermark;
}

/** 
 * Returns whether the watermark has to be embedded using a fast algorithm.
 *
 * @return Whether the watermark has to be embedded using a fast algorithm.
 */
const bool MajorityWatermark::getUseFastVersion() const {
	return useFastVersion;
}


/**
 * Returns the watermark bit located at index <tt>index</tt>.
 *
 * @param index The index of the watermark bit to retrieve.
 * @return The watermark bit located at index <tt>index</tt>.
 */
const bool MajorityWatermark::getBit(unsigned int index) const {
	return (watermarkBits[index/8]&(1<<(index%8)));
}

/**
 * Sets the watermark bit located at index <tt>index</tt>.
 *
 * @param index The index of the watermark bit to set.
 * @param value The new value for the watermark bit specified by <tt>index</tt>.
 */
void MajorityWatermark::setBit(unsigned int index, bool value) {
	if (value) {
		watermarkBits[index/8]|=(1<<(index%8));
	} else {
		watermarkBits[index/8]|=(1<<(index%8));
		watermarkBits[index/8]^=(1<<(index%8));
	}
}

/** 
 * Casts the watermark bit to a char pointer.
 */
MajorityWatermark::operator unsigned char*() {
	return watermarkBits;
}

/** 
 * Returns the module name of this class.
 *
 * @return "Majority Watermark";
 */
const char* MajorityWatermark::getModuleName() const {
	return "Majority Watermark";
} 

/** 
 * Returns the module description of this class.
 *
 * @return The module description of this class.
 */
const char* MajorityWatermark::getModuleDescription() const {
	return "Implements a majority watermark (each bit is embedded multiple times).";
} 

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
