//
// Title:   NearestNeighbourProjection.cpp
// Created: November 28, 2002
// Authors: Daniel Cotting
//
// Copyright (c) 2001, 2002, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.
//
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

/** 
 * @file NearestNeighbourProjection.cpp
 *
 * Projects a patch of the original object onto the surface of the verification object
 * by computing the center of mass of the k nearest neighbours of the point to project.
 */

#include <math.h>

#include "../../../Core/Scene/src/Scene.h"
#include "../../../Core/DataStructures/src/NeighbourHood.h"
#include "../../../Core/Interfaces/src/SurfelInterface.h"

#include "MySurfelList.h"
#include "NearestNeighbourProjection.h"
#include "Patch.h"

/**
 * The number of neighbours to consider when calculating the center of mass.
 */
#define numberOfNeighbours 1

/**
 * Projects the patch <tt>originalSurfelPatch</tt> onto the surface of the verification 
 * object defined by <tt>verificationSurfelList</tt>.
 *
 * @param originalSurfelPatch The patch to project.
 * @param verificationSurfelList The surface to project the patch onto.
 * @return The projected patch.
 */
SurfelCollection* NearestNeighbourProjection::projectPatch(
	Patch* originalSurfelPatch, 
	MySurfelList* verificationSurfelList) 
{
	
	// the following code projects each surfel of the original object onto a plane defined
	// by the three nearest neighbours of the verification surface.
	int i;
	int j;
	SurfelInterface* surfel;

	// create data structures for the neighbourhood search on the verification surface
	int nofVerificationSurfels = verificationSurfelList->count();
	Vector3D* verificationPositions = new Vector3D[nofVerificationSurfels];
	for (i = 0, surfel = verificationSurfelList->first(); i < nofVerificationSurfels; i++, surfel = verificationSurfelList->next()) {
		verificationPositions[i] = surfel->getPosition();
	}

	// initialize the neighbourhood search objects
	NeighbourHood* neighbourHoodOnVerificationSurface = new NeighbourHood(verificationPositions, nofVerificationSurfels);

	// create data structure which will hold the result
	SurfelCollection* projectedPatch = new SurfelCollection();
	SurfelInterface* projectedSurfel;

	// iterate over all the surfels of the original object and project them onto the modified surface
	int nofOriginalSurfels = originalSurfelPatch->count();
	for (i=0, surfel = originalSurfelPatch->first(); i < nofOriginalSurfels; i++, surfel = originalSurfelPatch->next()) {

		// set the source point to the position of the surfel to project.
		neighbourHoodOnVerificationSurface->setSourcePoint(surfel->getPosition());
		// get the nearest three neighbours
		neighbourHoodOnVerificationSurface->setNofNeighbours(numberOfNeighbours);

		Vector3D projectedPosition(0,0,0);
		for (j=0; j<numberOfNeighbours; j++) {
			projectedPosition += neighbourHoodOnVerificationSurface->getNeighbourPosition(j);
		}
		projectedPosition /= numberOfNeighbours;

		// create new surfel object
		// add the projected point to the patch
		projectedSurfel = projectedPatch->addSurfel(surfel, false);

		// modify the Position. this can be accomplished by using the scene even though 
		// the surfel is not really displayed
		Scene::getInstance()->setSurfelPosition(projectedSurfel, projectedPosition, false);
	}

	// clean up
	delete neighbourHoodOnVerificationSurface;
	delete verificationPositions;

	// return the result
	return projectedPatch;
}

/**
 * Returns the name of this module.
 *
 * @return The name of this module.
 */
const char* NearestNeighbourProjection::getModuleName() const {
	return "Nearest Neighbour Projection";
}

/**
 * Returns the description of this module.
 *
 * @return The description of this module.
 */
const char* NearestNeighbourProjection::getModuleDescription() const {
	return "Projects points of one patch onto the other surface by computing the center of mass of the k nearest neighbours.";
}

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
