//
// Title:   PCAPreprocessing.cpp
// Created: November 28, 2002
// Authors: Daniel Cotting
//
// Copyright (c) 2001, 2002, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.
//
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

/**
 * @file PCAPreprocessing.cpp
 *
 * Contains code for the preprocessing step of watermarking (like object reorientation and feature extraction).
 *
 * @author Daniel Cotting
 */

#include <qobject.h>
#include <qmessagebox.h>
#include <stdlib.h>

#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_eigen.h>

#include "../../../Core/Interfaces/src/SurfelInterface.h"
#include "../../../Core/Scene/src/Scene.h"

#include "MySurfelList.h"
#include "PatchBuilderInterface.h"
#include "PCAPreprocessing.h"
#include "PCA.h"
#include "WatermarkInterface.h"

/**
 * Constructor of the <tt>PCAPreprocessing</tt> class.
 * Only initializes the random number generator to a predefined seed.
 */
PCAPreprocessing::PCAPreprocessing(): random(Random(3615)) {
}

/**
 * Destructor of the <tt>PCAPreprocessing</tt> class.
 */
PCAPreprocessing::~PCAPreprocessing() {
}

/** 
 * Preprocesses the surfels contained in the list. This step is performed 
 * before embedding the watermark in order to get a deterministic reorientation 
 * of the object in space.
 *
 * @param surfelList The list of surfels to preprocess and reorient in space.
 * @param scale The scale to use for the geometry.
 */
void PCAPreprocessing::preprocess(MySurfelList* surfelList, Vector3D scale) {

	// used for coordinate flipping/mirroring (when user requests it)
	bool flipX = false;
	bool flipY = false;
	bool flipZ = false;

	// the user decides when to leave the loop
	bool doOneMoreIteration = true;
	while (doOneMoreIteration) {

		// for now decide not to iterate ony more. the user changes this variable later on...
		doOneMoreIteration = false;

		// allocate matrices to compute PCA over nofSurfels points of the list
		unsigned int nofSurfels;
		nofSurfels = surfelList->count(); 

		// calculate the dimension of the matrices
		unsigned int matrixRows = nofSurfels; 
		unsigned int matrixCols = 3;

		// allocate the matrices
		gsl_matrix *dataMatrix = gsl_matrix_alloc(matrixRows, matrixCols);
		gsl_matrix *covarianceMatrix = gsl_matrix_alloc(matrixCols, matrixCols);

		// create the initial matrix and at the same time compute the center of mass
		// first initialze some data
		unsigned int surfelNumber = 0;								// no surfels added to the matrix yet
		double centerX = 0.0;										// center not yet calculated
		double centerY = 0.0;
		double centerZ = 0.0;

		// add nofSurfels random points to the matrix
		while (surfelNumber < nofSurfels) {
			for (SurfelInterface* surfel = surfelList->first(); surfel!=0; surfel=surfelList->next()) {
				Vector3D position = surfel->getPosition();
				// add surfel coordinates to the matrix
				gsl_matrix_set(dataMatrix, surfelNumber, 0, (double)position[0]);
				gsl_matrix_set(dataMatrix, surfelNumber, 1, (double)position[1]);
				gsl_matrix_set(dataMatrix, surfelNumber, 2, (double)position[2]);
				// update the center of mass temporary value
				centerX += position[0];
				centerY += position[1];
				centerZ += position[2];
				surfelNumber++;
			}
		}

		// calculate the final center of mass
		centerX /= surfelNumber;
		centerY /= surfelNumber;
		centerZ /= surfelNumber;

		// compute the PCA
		PCA pca;
		pca.createCovarianceMatrix(dataMatrix, covarianceMatrix);
		PCA::EigenList eigenList = pca.initializeEigenList(matrixCols);
		pca.calculateEigenSystem(covarianceMatrix, eigenList ); 

		// Extract the eigenvectors (first one is the one corresponding to the largest eigenvalue)
		eigenList = eigenList->next;
		float eigenvector1X = gsl_vector_get(eigenList->eigenvector, 0);
		float eigenvector1Y = gsl_vector_get(eigenList->eigenvector, 1);
		float eigenvector1Z = gsl_vector_get(eigenList->eigenvector, 2);
		float eigenvalue1Rescale = 1.0f/sqrt(eigenList->eigenvalue);
		// qDebug("Eigenvalue1: %f", eigenList->eigenvalue);
		eigenList = eigenList->next;
		float eigenvector2X = gsl_vector_get(eigenList->eigenvector, 0);
		float eigenvector2Y = gsl_vector_get(eigenList->eigenvector, 1);
		float eigenvector2Z = gsl_vector_get(eigenList->eigenvector, 2);
		float eigenvalue2Rescale = 1.0f/sqrt(eigenList->eigenvalue);
		// qDebug("Eigenvalue2: %f", eigenList->eigenvalue);
		eigenList = eigenList->next;
		float eigenvector3X = gsl_vector_get(eigenList->eigenvector, 0);
		float eigenvector3Y = gsl_vector_get(eigenList->eigenvector, 1);
		float eigenvector3Z = gsl_vector_get(eigenList->eigenvector, 2);
		float eigenvalue3Rescale = 1.0f/sqrt(eigenList->eigenvalue); 
		// qDebug("Eigenvalue3: %f", eigenList->eigenvalue);

		// free memory
		PCA::EigenList eigenHead = eigenList;
		eigenList=eigenList->next;
		while (eigenList!=NULL) {
			PCA::EigenList nextElement = eigenList->next;
			free(eigenList->annotation);
			gsl_vector_free(eigenList->eigenvector);
			free(eigenList);
			eigenList = nextElement;
		}
		free(eigenHead);

		// free matrices
		gsl_matrix_free(dataMatrix);
		gsl_matrix_free(covarianceMatrix);

		// flip coordinates if required and reset the flip variables
		if (flipX) {
			eigenvector1X-=eigenvector1X;
			eigenvector1Y-=eigenvector1Y;
			eigenvector1Z-=eigenvector1Z;
		}
		if (flipY) {
			eigenvector2X-=eigenvector2X;
			eigenvector2Y-=eigenvector2Y;
			eigenvector2Z-=eigenvector2Z;
		}
		if (flipZ) {
			eigenvector3X-=eigenvector3X;
			eigenvector3Y-=eigenvector3Y;
			eigenvector3Z-=eigenvector3Z;
		}
		flipX = false;
		flipY = false;
		flipZ = false;

		// check whether the user wants to rescale the geometry
		if (scale[0]!=0 && scale[1]!=0 && scale[2]!=0) {
			// adjust the rescaling values
			eigenvalue1Rescale *= sqrt(scale[0]); 
			eigenvalue2Rescale *= sqrt(scale[1]); 
			eigenvalue3Rescale *= sqrt(scale[2]); 
		} else {
			// the following three lines have to be commented if one wants uniform rescaling
			// in that case multiply all values by 50 or the model will be too small,
			// i.e. 
			// eigenvalue1Rescale *= 50; eigenvalue2Rescale *= 50; eigenvalue3Rescale *= 50;
			eigenvalue1Rescale = 1.0f; 
			eigenvalue2Rescale = 1.0f; 
			eigenvalue3Rescale = 1.0f; 
			// rescaling is cleared now. we won't change the scale...
		}

		// calculate matrix to transform the normal
		// first get matrix
		float m11 = eigenvector1X*eigenvalue1Rescale;
		float m12 = eigenvector1Y*eigenvalue1Rescale;
		float m13 = eigenvector1Z*eigenvalue1Rescale;
		float m21 = eigenvector2X*eigenvalue2Rescale;
		float m22 = eigenvector2Y*eigenvalue2Rescale;
		float m23 = eigenvector2Z*eigenvalue2Rescale;
		float m31 = eigenvector3X*eigenvalue3Rescale;
		float m32 = eigenvector3Y*eigenvalue3Rescale;
		float m33 = eigenvector3Z*eigenvalue3Rescale;
		// then get the inverse
		float detInverse = 1/(m11*(m33*m22-m32*m23)-m21*(m33*m12-m32*m13)+m31*(m23*m12-m22*m13));
		float i11 = (m33*m22-m32*m23)*detInverse;  
		float i12 =-(m33*m12-m32*m13)*detInverse;    
		float i13 = (m23*m12-m22*m13)*detInverse;   
		float i21 =-(m33*m21-m31*m23)*detInverse; 
		float i22 = (m33*m11-m31*m13)*detInverse;   
		float i23 =-(m23*m11-m21*m13)*detInverse;  
		float i31 = (m32*m21-m31*m22)*detInverse;   
		float i32 =-(m32*m11-m31*m12)*detInverse;    
		float i33 = (m22*m11-m21*m12)*detInverse;   
		// and then transpose
		float t11 = i11;
		float t12 = i21;
		float t13 = i31;
		float t21 = i12;
		float t22 = i22;
		float t23 = i32;
		float t31 = i13;
		float t32 = i23;
		float t33 = i33;
 
		// performs a transformation on all the points
		Scene* scene = Scene::getInstance();
		for (SurfelInterface* surfel = surfelList->first(); surfel!=0; surfel=surfelList->next()) {

			// transform the position
			Vector3D position = surfel->getPosition();
			float p0 = position[0] - centerX;
			float p1 = position[1] - centerY;
			float p2 = position[2] - centerZ;
			float newX = p0*m11+p1*m12+p2*m13;
			float newY = p0*m21+p1*m22+p2*m23;
			float newZ = p0*m31+p1*m32+p2*m33;
			// rescale
			position[0] = newX;
			position[1] = newY;
			position[2] = newZ;
			scene->setSurfelPosition(surfel, position, false);
		  
			// transform the normal
			Vector3D normal = surfel->getNormal();
			newX = normal[0]*t11+normal[1]*t12+normal[2]*t13;
			newY = normal[0]*t21+normal[1]*t22+normal[2]*t23;
			newZ = normal[0]*t31+normal[1]*t32+normal[2]*t33;
			// calculate the rescaling factor (normals have to be of size 1)
			float factor = 1/sqrt(newX*newX+newY*newY+newZ*newZ);
			normal[0] = newX*factor;
			normal[1] = newY*factor;
			normal[2] = newZ*factor;
			scene->setSurfelNormal(surfel, normal, false);
		}

		// we emit the signal so that the screen gets updated
		surfel = surfelList->first();
		scene->setSurfelPosition (surfel, surfel->getPosition(), true);

		// prepare the message that will be shown to the user
		char message[500];
		sprintf(message, "PCA Reorientation Done.\n\n  %02.5f %02.5f %02.5f\n  %02.5f %02.5f %02.5f\n%  02.5f %02.5f %02.5f\n\n Do you want to continue iterating?",
			eigenvector1X, eigenvector1Y, eigenvector1Z,
			eigenvector2X, eigenvector2Y, eigenvector2Z,
			eigenvector3X, eigenvector3Y, eigenvector3Z);

		// now ask the user and assign the objects based on the response
		switch( QMessageBox::information( NULL, "Question",	message,
									"Yes", "No", 0,
									 0, 1 ))
		{
			case 0: 
				doOneMoreIteration = true;
				if (!QMessageBox::information( NULL, "Question", "Do you want to flip X coordinate?",
											"Yes", "No", 0, 0, 1 ))	{
					flipX = true;
				}
				if (!QMessageBox::information( NULL, "Question", "Do you want to flip Y coordinate?",
											"Yes", "No", 0, 0, 1 ))	{
					flipY = true;
				}
				if (!QMessageBox::information( NULL, "Question", "Do you want to flip Z coordinate?",
											"Yes", "No", 0, 0, 1 ))	{
					flipZ = true;
				}
				break;
			case 1:
				doOneMoreIteration = false;
				break;
		}

		// the aim of the following code was to visually undo the previous changes without changing the 
		// position of the points explicitly.
		// due to a bug in PointShop3D the code could not be used.
		// note that the code might be in an incorrect state because of the experiments last performed with it.
		/*
			// change the transformation matrix of the entire object, so 
			// that the user does not recognize any changes. the user wants the signed object to 
			// look like the original one (he doesn't want any stretched one).

			// first undo the scaling

			// scene->setScale(1/eigenvalue1InvSqrt, 1/eigenvalue2InvSqrt, 1/eigenvalue3InvSqrt);
			//	or
			scene->getActiveObject()->scale(1/eigenvalue1InvSqrt, 1/eigenvalue2InvSqrt, 1/eigenvalue3InvSqrt);

			// performs a transformation on all the points
			for (surfel = surfelList->first(); surfel!=0; surfel=surfelList->next()) {
				// transform the normal
				Vector3D normal = surfel->getNormal();
				float newX = normal[0]*eigenvalue1InvSqrt;//*eigenvalue1InvSqrt;
				float newY = normal[1]*eigenvalue2InvSqrt;//*eigenvalue2InvSqrt;
				float newZ = normal[2]*eigenvalue3InvSqrt;//*eigenvalue3InvSqrt;
				float factor = 1.0f/sqrt(newX*newX+newY*newY+newZ*newZ);
				normal[0] = newX*factor*eigenvalue1InvSqrt;
				normal[1] = newY*factor*eigenvalue2InvSqrt; 
				normal[2] = newZ*factor*eigenvalue3InvSqrt;
				scene->setSurfelNormal(surfel, normal, false);
			}
		 
			// we emit the signal so that the screen gets updated
			surfel = surfelList->first();
			scene->setSurfelPosition (surfel, surfel->getPosition(), true);
			QMessageBox::information( NULL, "Information", "Scaling Done. Next Step, Rotation..." );

			// then undo the rotation
			MyDataTypes::TransformationMatrix16f matrix;
			scene->getActiveObject()->getRotationMatrix(matrix);

			// we now have to multiply the rotation matrix with the inverse of our transformation matrix 
			// to undo the changes
			MyDataTypes::TransformationMatrix16f inverseTransformation = {
					i11, i12, i13, 0,
					i21, i22, i23, 0,
					i31, i32, i33, 0,
					0,   0,   0, 1
			};

			MyDataTypes::TransformationMatrix16f newMatrix;
			multiplyMatrix(newMatrix, matrix, inverseTransformation);
			scene->getActiveObject()->setRotationMatrix(newMatrix);
			QMessageBox::information( NULL, "Information", "Rotation Done." );
		*/
	}
}

/**
 * Multiplies a 4x4 matrix with another one of the same dimension.
 *
 * @param resultMatrix The matrix to store the result in.
 * @param matrix1 The matrix to multiply with the second one.
 * @param matrix2 The matrix to multiply with the first one.
 */
void PCAPreprocessing::multiplyMatrix(
		MyDataTypes::TransformationMatrix16f resultMatrix,
		MyDataTypes::TransformationMatrix16f matrix1,
		MyDataTypes::TransformationMatrix16f matrix2) 
{
	for (int row=0; row<4; row++) {
		for (int col=0; col<4; col++) {
			(resultMatrix)[col*4+row] = 0;
			for (int i=0; i<4; i++) {
				(resultMatrix)[col*4+row] += matrix1[col*4+i]*matrix2[i*4+row];
			}	
		}
	}
}

/**
 * Extracts a feature of the surfels contained in the list passed to this function.
 * A feature is a specific property of a point cloud geometry and is used to calculate
 * a watermark which is dependent on the object to be watermarked. In this implementation
 * the feature is calculated using the eigenvectors and eigenvalues of the points of the list.
 *
 * @param surfelList The list of surfels for which the feature has to be calculated.
 * @return A zero-terminated string dependent on the eigensystem of the points of the list.
 */
char* PCAPreprocessing::extractFeature(MySurfelList* surfelList) {
	
	// allocate matrices to compute PCA over nofSurfels points of the list
	unsigned int nofSurfels;
	nofSurfels = surfelList->count(); 

	// calculate the dimension of the matrices
	unsigned int matrixRows = nofSurfels; 
	unsigned int matrixCols = 3;
	gsl_matrix *a = gsl_matrix_alloc(matrixRows, matrixCols);
	gsl_matrix *s = gsl_matrix_alloc(matrixCols, matrixCols);


	// add all points to the matrix
	unsigned int surfelNumber = 0;							// number of surfels added to the matrix

	// iterate over all surfels
	for (SurfelInterface* surfel = surfelList->first(); surfel!=0; surfel=surfelList->next()) {
		// add surfel position to the matrix
		Vector3D position = surfel->getPosition();
		gsl_matrix_set( a, surfelNumber, 0, (double)position[0]);
		gsl_matrix_set( a, surfelNumber, 1, (double)position[1]);
		gsl_matrix_set( a, surfelNumber, 2, (double)position[2]);
		surfelNumber++;
	}

	// compute the PCA
	PCA pca;
	pca.createCovarianceMatrix(a, s);
	PCA::EigenList eigenList = pca.initializeEigenList(matrixCols);
	pca.calculateEigenSystem(s, eigenList ); 

	// Extract the eigenvectors (first one is the one corresponding to the largest eigenvalue)
	eigenList = eigenList->next;
	float eigenvector1X = gsl_vector_get(eigenList->eigenvector, 0);
	float eigenvector1Y = gsl_vector_get(eigenList->eigenvector, 1);
	float eigenvector1Z = gsl_vector_get(eigenList->eigenvector, 2);
	float eigenvalue1 = eigenList->eigenvalue;
	eigenList = eigenList->next;
	float eigenvector2X = gsl_vector_get(eigenList->eigenvector, 0);
	float eigenvector2Y = gsl_vector_get(eigenList->eigenvector, 1);
	float eigenvector2Z = gsl_vector_get(eigenList->eigenvector, 2);
	float eigenvalue2 = eigenList->eigenvalue;
	eigenList = eigenList->next;
	float eigenvector3X = gsl_vector_get(eigenList->eigenvector, 0);
	float eigenvector3Y = gsl_vector_get(eigenList->eigenvector, 1);
	float eigenvector3Z = gsl_vector_get(eigenList->eigenvector, 2);
	float eigenvalue3 = eigenList->eigenvalue;

	// free memory
	PCA::EigenList eigenHead = eigenList;
	eigenList=eigenList->next;
	while (eigenList!=NULL) {
		PCA::EigenList nextElement = eigenList->next;
		free(eigenList->annotation);
		gsl_vector_free(eigenList->eigenvector);
		free(eigenList);
		eigenList = nextElement;
	}
	free(eigenHead);

	// free matrices
	gsl_matrix_free(a);
	gsl_matrix_free(s);

	char* feature = (char *)malloc(250 * sizeof(char)); // allocate enough memory for the features
	sprintf(feature, "%d/%d/%d/%01.6d/%01.6d/%01.6d/%01.6d/%01.6d/%01.6d/%01.6d/%01.6d/%01.6d",
			eigenvalue1, eigenvalue2, eigenvalue3,
			eigenvector1X, eigenvector1Y,eigenvector1Z,
			eigenvector2X,eigenvector2Y,eigenvector2Z,
			eigenvector3X, eigenvector3Y,eigenvector3Z);
	return feature;
}

/**
 * Gets the scale of the geometry described by the list of surfels. 
 *
 * @param surfelList The list containing the surfels whose scale has to be determined.
 * @return	A vector containing the scaling of the geometry. The first component is
 *			relative to the direction of the eigenvector of the largest eigenvalue.
 *			The second one relative to the eigenvector of the second largest eigenvalue.
 *			The third one relative to the eigenvector of the third largest eigenvalue.
 */
Vector3D PCAPreprocessing::getScale(MySurfelList* surfelList) {

	// allocate matrices to compute PCA over nofSurfels points of the list
	unsigned int nofSurfels;
	nofSurfels = surfelList->count(); 

	// calculate the dimension of the matrices
	unsigned int matrixRows = nofSurfels; 
	unsigned int matrixCols = 3;
	gsl_matrix *a = gsl_matrix_alloc(matrixRows, matrixCols);
	gsl_matrix *s = gsl_matrix_alloc(matrixCols, matrixCols);


	// add all points to the matrix
	unsigned int surfelNumber = 0;							// number of surfels added to the matrix

	// iterate over all surfels
	for (SurfelInterface* surfel = surfelList->first(); surfel!=0; surfel=surfelList->next()) {
		// add surfel position to the matrix
		Vector3D position = surfel->getPosition();
		gsl_matrix_set( a, surfelNumber, 0, (double)position[0]);
		gsl_matrix_set( a, surfelNumber, 1, (double)position[1]);
		gsl_matrix_set( a, surfelNumber, 2, (double)position[2]);
		surfelNumber++;
	}

	// compute the PCA
	PCA pca;
	pca.createCovarianceMatrix(a, s);
	PCA::EigenList eigenList = pca.initializeEigenList(matrixCols);
	pca.calculateEigenSystem(s, eigenList ); 

	// Extract the eigenvectors (first one is the one corresponding to the largest eigenvalue)
	eigenList = eigenList->next;
	float eigenvalue1 = eigenList->eigenvalue;
	eigenList = eigenList->next;
	float eigenvalue2 = eigenList->eigenvalue;
	eigenList = eigenList->next;
	float eigenvalue3 = eigenList->eigenvalue;
	Vector3D scale(eigenvalue1, eigenvalue2, eigenvalue3);

	// free memory
	PCA::EigenList eigenHead = eigenList;
	eigenList=eigenList->next;
	while (eigenList!=NULL) {
		PCA::EigenList nextElement = eigenList->next;
		free(eigenList->annotation);
		gsl_vector_free(eigenList->eigenvector);
		free(eigenList);
		eigenList = nextElement;
	}
	free(eigenHead);

	// free matrices
	gsl_matrix_free(a);
	gsl_matrix_free(s);

	return scale;
}

/** 
 * Returns the name of this module.
 *
 * @return "PCA Preprocessing"
 */
const char* PCAPreprocessing::getModuleName() const {
	return "PCA Preprocessing";
}

/**
 * Returns the description of this module.
 *
 * @return The description of this module.
 */
const char* PCAPreprocessing::getModuleDescription() const {
	return "Performs an surfel reorientation based on the Principal Component Analysis.";
}

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
