//
// Title:   PCAPreprocessing.h
// Created: November 28, 2002
// Authors: Daniel Cotting
//
// Copyright (c) 2001, 2002, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.
//
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifndef __PCAPREPROCESSING_H_
#define __PCAPREPROCESSING_H_

#include <qobject.h>

#include "../../../Core/Interfaces/src/PluginInterface.h"
#include "../../../Core/Interfaces/src/SurfelInterface.h"
#include "MySurfelList.h"
#include "PreprocessingInterface.h"
#include "WatermarkInterface.h"


/**
 * Contains code for the preprocessing step of watermarking (like object reorientation and feature extraction).
 */
class PCAPreprocessing : public PreprocessingInterface {

public:
	
	/**
	 * Constructor of the <tt>PCAPreprocessing</tt> class.
	 * Only initializes the random number generator to a predefined seed.
	 */
	PCAPreprocessing();

	/**
	 * Destructor of the <tt>PCAPreprocessing</tt> class.
	 */
	~PCAPreprocessing();

	/** 
	 * Preprocesses the surfels contained in the list. This step is performed 
	 * before embedding the watermark in order to get a deterministic reorientation 
	 * of the object in space.
	 *
	 * @param surfelList The list of surfels to preprocess and reorient in space.
	 * @param scale			A vector containing the new scaling of the geometry. The first 
	 *						component is relative to the direction of the eigenvector of the
	 *						largest eigenvalue.	The second one relative to the eigenvector 
	 *						of the second largest eigenvalue. The third one relative to the 
	 *						eigenvector of the third largest eigenvalue.
	 */
	void preprocess(MySurfelList* surfelList, Vector3D scale = Vector3D(0,0,0));

	/**
	 * Extracts a feature of the surfels contained in the list passed to this function.
	 * A feature is a specific property of a point cloud geometry and is used to calculate
	 * a watermark which is dependent on the object to be watermarked. In this implementation
	 * the feature is calculated using the eigenvectors of 100 points of the list.
	 * Therefore the object requires a minimum of 100 points. In a final version of a 
	 * watermarking plugin all points of the object should be used for calculating the feature.
	 *
	 * @param surfelList The list of surfels for which the feature has to be calculated.
	 * @return A zero-terminated string dependent on the eigenvectors of 100 points of the list.
	 */
	char* extractFeature(MySurfelList* surfelList);

	/**
	 * Gets the scale of the geometry described by the list of surfels. 
	 *
	 * @param surfelList The list containing the surfels whose scale has to be determined.
	 * @return	A vector containing the scaling of the geometry. The first component is
	 *			relative to the direction of the eigenvector of the largest eigenvalue.
	 *			The second one relative to the eigenvector of the second largest eigenvalue.
	 *			The third one relative to the eigenvector of the third largest eigenvalue.
	 */
	Vector3D getScale(MySurfelList* surfelList);

	/** 
	 * Returns the name of this module.
	 *
	 * @return "PCA Preprocessing"
	 */
	const char* getModuleName() const;

	/**
	 * Returns the description of this module.
	 *
	 * @return The description of this module.
	 */
	const char* getModuleDescription() const;

	/**
	 * Multiplies a 4x4 matrix with another one of the same dimension.
	 *
	 * @param resultMatrix The matrix to store the result in.
	 * @param matrix1 The matrix to multiply with the second one.
	 * @param matrix2 The matrix to multiply with the first one.
	 */
	void multiplyMatrix(
		MyDataTypes::TransformationMatrix16f resultMatrix,
		MyDataTypes::TransformationMatrix16f matrix1,
		MyDataTypes::TransformationMatrix16f matrix2
	); 

private:

	/**
	 * Random number generator object. Will be initialized by the constructor.
	 */
	Random random;
};

#endif  //__PCAPREPROCESSING_H_

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
