//
// Title:   PCARegistration.cpp
// Created: November 28, 2002
// Authors: Daniel Cotting
//
// Copyright (c) 2001, 2002, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.
//
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

/**
 * @file PCARegistration.cpp
 *
 * Registers two surfel lists, that is both geometries are moved and rescaled to 
 * become as close as possible. In the ideal case both geometries will look more 
 * or less the same and will be located at the same position is space with the same
 * orientation. This is used as a preprocessing step before extracting the 
 * watermark in watermarking schemes using two objects (the original one and the
 * verification object) to get the watermark.
 * This implementation uses Principal Component Analysis (PCA) to perform that task.
 */

#include <stdlib.h>
#include <qobject.h>
#include <qmessagebox.h>

#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_eigen.h>

#include "../../../Core/Interfaces/src/SurfelInterface.h"
#include "../../../Core/Scene/src/Scene.h"

#include "MySurfelList.h"
#include "PatchBuilderInterface.h"
#include "PCARegistration.h"
#include "PCAPreprocessing.h"


/**
 * Constructor of the <tt>PCARegistration</tt> class.
 */
PCARegistration::PCARegistration() {
}


/**
 * Destructor of the <tt>PCARegistration</tt> class.
 */
PCARegistration::~PCARegistration() {
}


/**
 * Registers the two geometries <tt>originalSurfelList</tt> and <tt>verificationSurfelList</tt>
 * using Principal Component Analysis.
 *
 * @param originalSurfelList The list containing the surfels of the original object.
 * @param verificationSurfelList The list containing the surfels of the verification object.
 */
void PCARegistration::registerObjects(
		MySurfelList* originalSurfelList, 
		MySurfelList* verificationSurfelList) 
{
	PCAPreprocessing preprocessing;
	preprocessing.preprocess(originalSurfelList);
	switch( QMessageBox::information( NULL, "Question",	"Do you want to try to rescale the verification object?",
								"Yes", "No", 0, 0, 1 ))	{
	case 1:
		// the user does not want to rescale the verification object
		preprocessing.preprocess(verificationSurfelList);
		break;
	case 0:
		// rescale the verification object
		Vector3D originalScale = preprocessing.getScale(originalSurfelList);
		preprocessing.preprocess(verificationSurfelList, originalScale);
		break;
	}
}

/**
 * Returns the name of this module.
 *
 * @return The name of this module.
 */
const char* PCARegistration::getModuleName() const {
	return "PCA Registration";
}

/**
 * Returns the description of this module.
 *
 * @return The description of this module.
 */
const char* PCARegistration::getModuleDescription() const {
	return "Registers two objects by performing PCA based reorientation and scaling on both objects.";
}

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
