//
// Title:   PublicWatermarking.h
// Created: November 28, 2002
// Authors: Daniel Cotting
//
// Copyright (c) 2001, 2002, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.
//
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifndef __PUBLICWATERMARKING_H_
#define __PUBLICWATERMARKING_H_

#include <qobject.h>
#include "WatermarkingInterface.h"
#include "PatchBuilderInterface.h"
#include "Random.h"
#include "WatermarkInterface.h"

/**
 * Implements a public watermarking scheme where the original object is no longer 
 * required when one wants to extract the watermark. This code has been implemented
 * for testing purposes. Experiments have shown that the approach chosen in this
 * file is currently not suited for real world applications because of the 
 * verly long runtime. The extraction is based on correlation calculation which 
 * only performs well for large watermarks with a few thousands of bits. 
 * To embed so many bits the frequency transformation requires a huge amount of 
 * time.
 * For the reasons mentioned above this code has not been completed and represents
 * more a work in progress.
 */
class PublicWatermarking: public WatermarkingInterface {

public:

	/** 
	 * Constructor of class <tt>PublicWatermarking</tt>.
	 */
	PublicWatermarking():random(new Random(3615)) {};

	/**
	 * Embeds a watermark into the currently selected surfels or the active object.
	 * The watermark is based on a passphrase, which is passed to this function.
	 *
	 * @param passPhrase	The secret, user defined passphrase to use for the watermark creation.
	 * @param watermark		A watermark configured with the options the user specified. Note that the 
	 *						watermark will be filled with content later on.
	 */
	void embedWatermark(const QString& passPhrase, WatermarkInterface* watermark);

	/**
	 * Extracts the public watermark by using the test geometry.
	 * Since embedding is dependent on a passphrase, the same passphrase has to be used for extraction.
	 *
	 * @param passPhrase	The passphrase to be used for extraction.
	 * @param watermark		A watermark configured with the options the user specified. Note that the 
	 *						watermark will be filled with content later on.
	 */
	void extractWatermark(const QString& passPhrase, WatermarkInterface* watermark);

	/**
	 * Returns the name of this module.
	 *
	 * @return The name of this module.
	 */
	const char* getModuleName() const;

	/**
	 * Returns the description of this module.
	 *
	 * @return The description of this module.
	 */
	const char* getModuleDescription() const;

protected:

	/**
	 * Embeds the watermark into the patch.
	 *
	 * @param patch The patch to embed the watermark into.
	 * @param watermark The watermark to embed into the patch. This object also contains the strength 
	 *					of the watermark (amplitude of spectral modifications while embedding). 
	 */
	void embedWatermarkIntoPatch(Patch* patch, WatermarkInterface* watermark);

	/**
	 * Extracts watermark from a patch. In fact, no watermark is explicitly extracted, but a correlatiion and a threshold value
	 * is calculated.
	 *
	 * @param patch						The patch to be processed.
	 * @param watermark					The watermark that is supposed to be contained in the patch
	 * @param correlation				This will contain the resulting correlation.
	 * @param threshold					This will contain the resulting threshold.
	 */
	void extractWatermarkFromPatch(Patch* patch, WatermarkInterface* watermark, float* correlation, float* threshold);

	/** 
	 * Random number generator.
	 */
	Random* random;
};

#endif __PUBLICWATERMARKING_H_

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
