//
// Title:   WatermarkDialogImpl.h
// Created: November 28, 2002
// Authors: Daniel Cotting
//
// Copyright (c) 2001, 2002, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.
//
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifndef WATERMARKDIALOGIMPL_H
#define WATERMARKDIALOGIMPL_H

#include "../UI/WatermarkDialog.h"
#include "WatermarkingInterface.h"

/**
 * Dialog box displaying a selection of various watermarking algorithms from which the user can select one.
 * The user can also specify a passphrase.
 */
class WatermarkDialogImpl : public WatermarkDialog
{ 
  Q_OBJECT

public:
	/**
	 * Constructor for the dialog of the watermarking plugin.
	 *
	 * @param algorithms					Array of algorithms to display in the dialog box. 
	 * @param parent						Standard QT dialog parameter.
	 * @param name							Standard QT dialog parameter.
	 * @param modal							Standart QT dialog parameter.
	 * @param f								Standard QT dialog parameter.
	 */
    WatermarkDialogImpl(WatermarkingInterface** watermarkingAlgorithms, QWidget* parent = 0, const char* name = 0, bool modal = FALSE, WFlags f = 0 );

public slots:

	/**
	 * Embeds the watermark by calling the appropriate function of the currently selected algorithm.
	 * A watermark configured with the parameters the user has specified is passed to the watermarking
	 * procedure.
	 */
	virtual void embedWatermark();

	/**
	 * Extracts the watermark by calling the appropriate function of the currently selected algorithm.
	 * A watermark configured with the parameters the user has specified is passed to the watermarking
	 * procedure.
	 */
	virtual void extractWatermark();

	/**
	 * Sets the passphrase to the specified value.
	 *
	 * @param passphraseIn The new passphrase to use for the watermark.
	 */
	virtual void setPassphrase(const QString& passphrase);

	/**
	 * Sets the watermark bit length to the specified value.
	 *
	 * @param watermarkBitLengthIn The new watermark length value to use (in bits).
	 */
	virtual void setWatermarkBitLength(int watermarkBitLengthIn);


	/**
	 * Sets the watermark chip rate to the specified value.
	 *
	 * @param watermarkChipRateIn The new watermark chip rate to use (number of repetitions).
	 */
	virtual void setWatermarkChipRate(int watermarkChipRateIn);

	/**
	 * Sets the watermark usage divider to the specified value.
	 *
	 * @param watermarkUsageDividerIn The new watermark usage divider to use (fraction of frequencies to use for embedding).
	 */
	virtual void setWatermarkUsageDivider(int watermarkUsageDividerIn);

	/**
	 * Sets the watermark strength (alpha value) to the specified value.
	 *
	 * @param watermarkStrength The new watermark strength value to use (in promille).
	 */
	virtual void setWatermarkStrengthInPromille(int watermarkStrengthInPromille);

	/**
	 * Sets the selected algorithm to the algorithm having the same name as the passed parameter.
	 *
	 * @param algorithmIn The name of the algorithm to select.
	 */
    virtual void setAlgorithm(const QString& algorithm);

	/**
	 * Sets whether the watermark has to be applied to color or position.
	 *
	 * @param colorWatermarkIn Defines whether the watermark has to be applied to the color attribute.
	 */
	virtual void setIsColorWatermark(int colorWatermarkIn);

	/**
	 * Sets whether the fast version of the embedding is to be used.
	 *
	 * @param selectionIndex Index of the user selection (0 for fast, 1 for accurate).
	 */
	virtual void setUseFastVersion(int selectionIndex);


private:
	QString& passphrase;									///< passphrase that user has entered in text field of UI
	int watermarkBitLength;									///< length of the watermark in bits
	int watermarkChipRate;									///< chip rate of the watermark
	int watermarkUsageDivider;								///< usage divider of the watermark
	WatermarkingInterface** watermarkingAlgorithms;			///< array of watermarking algorithms which are available
	WatermarkingInterface* selectedWatermarkingAlgorithm;	///< the currently selected watermarking algorithm
	float watermarkStrength;								///< the strength of the watermark
	bool colorWatermark;									///< Is the watermark a watermark on color.
	bool useFastVersion;									///< Do we want to use the fast embedding algorithm?
};

#endif


// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
