/**
 * @file dsaupd.h
 *
 * Header file for code to calculate the eigenvalues of a matrix using
 * a callback function and ARPACK solver.
 */

#include <math.h>

typedef void CallbackFunction(unsigned int n, double *in, double *out);

// function prototype for the call to ARPACK FORTRAN routine
extern "C" void dsaupd_(int *ido, char *bmat, int *n, char *which,
			int *nev, double *tol, double *resid, int *ncv,
			double *v, int *ldv, int *iparam, int *ipntr,
			double *workd, double *workl, int *lworkl,
			int *info);

// function prototype for the call to ARPACK FORTRAN routine
extern "C" void dseupd_(int *rvec, char *All, int *select, double *d,
			double *v, int *ldv, double *sigma, 
			char *bmat, int *n, char *which, int *nev,
			double *tol, double *resid, int *ncv, double *v2,
			int *ldv2, int *iparam, int *ipntr, double *workd,
			double *workl, int *lworkl, int *ierr);

/**
 * Simplified function call to the ARPACK solver routine for a 
 * simple, symmetric eigenvalue problem Av = lv. 
 *
 * The looping procedure and final extraction of eigenvalues and 
 * vectors is handled automatically. Most of the parameters to 
 * the FORTRAN functions are hidden from the user, since most of
 * them are determined from user input anyway.
 * To use the function call, there must be a function
 * defined in the calling program of the form
 *
 * <tt>av(int n, double *in, double *out)</tt>
 *
 * where the function av finds out = A.in, and n is the order of
 * the matrix A. It is used in the looping procedure.
 *
 * @param av	The callback function.
 * @param n		The order of the square matrix A.
 * @param nev	The number of eigenvalues to be found, starting at 
 *				the bottom. Note that the highest eigenvalues, or 
 *				some other choices, can be found. For now, the 
 *				choice of the lowest nev eigenvalues is hard-coded.
 * @param Evals	A one-dimensional array of length nev to hold the
 *				eigenvalues.
 * 
 * @author Scot Shaw
 * @author Daniel Cotting
 */
void dsaupd(CallbackFunction av, int n, int nev, double *Evals);

/**
 * Simplified function call to the ARPACK solver routine for a 
 * simple, symmetric eigenvalue problem Av = lv. 
 *
 * The looping procedure and final extraction of eigenvalues and 
 * vectors is handled automatically. Most of the parameters to 
 * the FORTRAN functions are hidden from the user, since most of
 * them are determined from user input anyway.
 * To use the function call, there must be a function
 * defined in the calling program of the form
 *
 * <tt>av(int n, double *in, double *out)</tt>
 *
 * where the function av finds out = A.in, and n is the order of
 * the matrix A. It is used in the looping procedure.
 *
 * @param av	The callback function.
 * @param n		The order of the square matrix A.
 * @param nev	The number of eigenvalues to be found, starting at 
 *				the bottom. Note that the highest eigenvalues, or 
 *				some other choices, can be found. For now, the 
 *				choice of the lowest nev eigenvalues is hard-coded.
 * @param Evals	A one-dimensional array of length nev to hold the
 *				eigenvalues.
 * @param Evecs	A two-dimensional array of size nev by n to hold the
 *				eigenvectors. Note that the	vectors are stored as 
 *				columns of Evecs, so that the elements of vector i
 *				are the values Evecs[j][i].
 */
void dsaupd(CallbackFunction av, int n, int nev, double *Evals, double **Evecs);





